#!/usr/bin/python
from os import system
import os
from subprocess import Popen

class Excursion(object):

    RFS = {}

    @classmethod
    def begin(cls, rfs):
        cls.RFS[rfs.path] = []

    @classmethod
    def add(cls, rfs, path, restore=True, dst=None):
        cls.RFS[rfs.path].append(Excursion(path, restore, dst))

    @classmethod
    def do(cls, rfs):
        r = cls.RFS[rfs.path]
        for tmp in r:
            tmp._do_excursion(rfs)

    @classmethod
    def end(cls, rfs):
        r = cls.RFS[rfs.path]
        for tmp in r:
            tmp._undo_excursion(rfs)
        del r

    def __init__(self, path, restore, dst):
        self.origin = path
        self.restore = restore
        self.dst = dst


    def _do_excursion(self, rfs):
        if rfs.lexists(self.origin) and self.restore is True:
            save_to = "%s.orig" % self.origin
            system('mv %s %s' % (rfs.fname(self.origin), rfs.fname(save_to)))
        if os.path.exists(self.origin):
            if self.dst is not None:
                dst = self.dst
            else:
                dst = self.origin
            system('cp %s %s' % (self.origin, rfs.fname(dst)))

    def _undo_excursion(self, rfs):
        saved_to = "%s.orig" % self.origin
        if rfs.lexists(self.origin):
            flags = "-f"
            if rfs.isdir(self.origin):
                flags += "r"
            system('rm %s %s' % (flags, rfs.fname(self.origin)))
        if self.restore is True and rfs.lexists(saved_to):
            system('mv %s %s' % (rfs.fname(saved_to), rfs.fname(self.origin)))

class Filesystem(object):

    # pylint: disable=too-many-public-methods

    def __init__(self, path, clean=False):
        self.path = os.path.abspath(path)

    def fname(self, path):
        if path.startswith('/'):
            path = path[1:]
        return os.path.join(self.path, path)

    def open(self, path, mode="r"):
        return open(self.fname(path), mode)

    def isdir(self, path):
        return os.path.isdir(self.fname(path))

    def islink(self, path):
        return os.path.islink(self.fname(path))

    def isfile(self, path):
        return os.path.isfile(self.fname(path))

    def exists(self, path):
        return os.path.exists(self.fname(path))

    def lexists(self, path):
        return os.path.lexists(self.fname(path))

    def mkdir(self, path):
        os.makedirs(self.fname(path))

    def realpath(self, path):

        path = path.split(os.sep)
        path.reverse()
        following = []
        real_path = [self.path]

        while path:
            candidate = path.pop()

            # Don't care
            if candidate == '' or candidate == os.curdir:
                continue

            # Can't go out of RFS
            if candidate == os.pardir:
                if following:
                    following.pop()
                if len(real_path) > 1:
                    real_path.pop()
                continue

            parent = os.sep.join(real_path)
            new_path = os.path.join(parent, candidate)
            if not os.path.islink(new_path):
                if following:
                    following.pop()
                real_path.append(candidate)
                continue

            # Circular loop; Don't follow it
            if new_path in following:
                real_path.append(candidate)
                continue

            following.append(new_path)
            link = os.readlink(new_path)

            # Reset root for absolute link
            if os.path.isabs(link):
                real_path = [self.path]

            for element in reversed(link.split(os.sep)):
                path.append(element)

        return os.sep.join(real_path)

    def symlink(self, src, path, allow_exists=False):
        try:
            os.symlink(src, self.fname(path))
        except OSError as e:
            if e.errno != os.errno.EEXIST:
                raise
            elif not allow_exists:
                raise

    def stat(self, path):
        return os.stat(self.fname(path))

    def chown(self, path, uid, gid):
        os.chown(self.fname(path), uid, gid)

    def chmod(self, path, mode):
        os.chmod(self.fname(path), mode)

    def utime(self, path, times=None):
        os.utime(self.fname(path), times)

    def cat_file(self, inf):
        content = []
        try:
            f = self.open(inf)
            content = f.readlines()
            f.close()
        except IOError:
            pass
        return content

    def remove(self, path, noerr=False):
        try:
            return os.remove(self.fname(path))
        except BaseException:
            if not noerr:
                raise

    def rmtree(self, path):
        shutil.rmtree(self.fname(path))

    def listdir(self, path='', ignore=None, skiplinks=False):
        if not ignore:
            ignore = []

        retval = [
            os.path.join(
                self.path,
                path,
                x) for x in os.listdir(
                self.fname(path)) if x not in ignore]

        if skiplinks:
            retval = [
                x for x in retval if (
                    not os.path.islink(x)) and os.path.isdir(x)]

        return retval

    def glob(self, path):
        flist = glob(self.fname(path))
        for i in flist:
            if not i.startswith(self.path):
                raise IOError("Broken glob '%s'" % path)

        return flist

    def write_file(self, path, mode, cont):
        f = self.open(path, "w")
        f.write(cont)
        f.close()
        if mode is not None:
            self.chmod(path, mode)

    def append_file(self, path, content, mode=None):
        f = self.open(path, "a")
        f.write(content)
        f.close()
        if mode is not None:
            self.chmod(path, mode)

    def read_file(self, path):
        fp = self.open(path, "r")
        retval = fp.read()
        fp.close()
        return retval

    def mkdir_p(self, newdir, mode=0o755):
        """works the way a good mkdir -p would...
                - already exists, silently complete
                - regular file in the way, raise an exception
                - parent directory(ies) does not exist, make them as well
        """
        if self.isdir(newdir):
            pass
        elif self.isfile(newdir):
            raise OSError(
                "a file with the same name as the desired "
                "dir, '%s', already exists. in RFS %s" %
                (newdir, self.path))
        else:
            self.mkdir(newdir)
            self.chmod(newdir, mode)

class ChRootFilesystem(Filesystem):

    def __init__(self, path, interpreter=None):
        Filesystem.__init__(self, path)
        self.interpreter = interpreter
        self.cwd = os.open("/", os.O_RDONLY)
        self.inchroot = False

    def __del__(self):
        os.close(self.cwd)

    def __enter__(self):
        Excursion.begin(self)
        Excursion.add(self, "/etc/resolv.conf")
        Excursion.add(self, "/etc/apt/apt.conf")
        Excursion.add(self, "/usr/sbin/policy-rc.d")

        if self.interpreter:
            if not self.exists("usr/bin"):
                if self.islink("usr/bin"):
                    Excursion.add(self, "/usr/bin")

            ui = "/usr/share/elbe/qemu-elbe/" + self.interpreter
            if not os.path.exists(ui):
                ui = "/usr/bin/" + self.interpreter

            Excursion.add(self, ui, False, "/usr/bin")

        Excursion.do(self)

        self.mkdir_p("usr/bin")
        self.mkdir_p("usr/sbin")
        self.write_file("usr/sbin/policy-rc.d", 0o755, "#!/bin/sh\nexit 101\n")
        self.mount()


    def __exit__(self, _typ, _value, _traceback):
        if self.inchroot:
            self.leave_chroot()
        self.umount()

        Excursion.end(self)

    def mount(self):
	if self.path == '/':
            return
        try:
            system("mount -t proc none %s/proc" % self.path)
            system("mount -t sysfs none %s/sys" % self.path)
            system("mount -o bind /dev %s/dev" % self.path)
            system("mount -o bind /dev/pts %s/dev/pts" % self.path)
        except BaseException:
            self.umount()
            raise

    def enter_chroot(self):
        assert not self.inchroot

        os.environ["LANG"] = "C"
        os.environ["LANGUAGE"] = "C"
        os.environ["LC_ALL"] = "C"

        os.chdir(self.path)
        self.inchroot = True

        if self.path == '/':
            return

        os.chroot(self.path)

    def _umount(self, path):
        if os.path.ismount(path):
            system("umount %s" % path)

    def umount(self):
        if self.path == '/':
            return
        self._umount("%s/proc/sys/fs/binfmt_misc" % self.path)
        self._umount("%s/proc" % self.path)
        self._umount("%s/sys" % self.path)
        self._umount("%s/dev/pts" % self.path)
        self._umount("%s/dev" % self.path)

    def leave_chroot(self):
        assert self.inchroot

        os.fchdir(self.cwd)

        self.inchroot = False
        if self.path == '/':
            return

        os.chroot(".")


def chroot(directory, cmd, **args):
    os.environ["LANG"] = "C"
    os.environ["LANGUAGE"] = "C"
    os.environ["LC_ALL"] = "C"
    chcmd = "chroot %s %s" % (directory, cmd)
    system(chcmd)

if __name__ == "__main__":
    rfs = ChRootFilesystem(os.sys.argv[1])
    with rfs:
        try:
            while True:
                cmd = raw_input("%s $ " % rfs.path)
                chroot(rfs.path, cmd)
        except EOFError:
            pass
